//
//  RGraphicsDevice.m
//  R
//
//  Created by Byron Ellis on 1/19/07.
//  Copyright 2007 __MyCompanyName__. All rights reserved.
//

#import "RGraphicsDevice.h"

#include <Rinternals.h>
#include <Rgraphics.h>
#include <Rdevices.h>

#include <R_ext/GraphicsDevice.h>
#include <R_ext/GraphicsEngine.h>



#pragma mark Graphics Device Prototypes
#pragma mark -

typedef struct {
	RGraphicsDevice *device;
	CGContextRef     ctx;
	double           tscale;
	BOOL             locator;
} QuartzDesc;


static Rboolean RQuartz_Open(NewDevDesc*,QuartzDesc*,char*,double,double,int);
static void     RQuartz_Close(NewDevDesc*);
static void     RQuartz_Activate(NewDevDesc*);
static void     RQuartz_Deactivate(NewDevDesc*);
static void     RQuartz_Size(double*,double*,double*,double*,NewDevDesc*);
static void     RQuartz_NewPage(R_GE_gcontext*,NewDevDesc*);
static void     RQuartz_Clip(double,double,double,double,NewDevDesc*);
static double   RQuartz_StrWidth(char*,R_GE_gcontext*,NewDevDesc*);
static void     RQuartz_Text(double,double,char*,double,double,R_GE_gcontext*,NewDevDesc*);
static void     RQuartz_Rect(double,double,double,double,R_GE_gcontext*,NewDevDesc*);
static void     RQuartz_Circle(double,double,double,R_GE_gcontext*,NewDevDesc*);
static void     RQuartz_Line(double,double,double,double,R_GE_gcontext*,NewDevDesc*);
static void     RQuartz_Polyline(int,double*,double*,R_GE_gcontext*,NewDevDesc*);
static void     RQuartz_Polygon(int,double*,double*,R_GE_gcontext*,NewDevDesc*);
static Rboolean RQuartz_Locator(double*,double*,NewDevDesc*);
static void     RQuartz_Mode(int mode,NewDevDesc*);
static void     RQuartz_Hold(NewDevDesc*);
static void     RQuartz_MetricInfo(int,R_GE_gcontext *,double*,double*,double*,NewDevDesc*);

#pragma mark Object Implementation
#pragma mark -

@implementation RGraphicsDevice
+ (RGraphicsDevice*)deviceWithContext:(CGContextRef)aContext deviceDesc:(void*)dd delegate:(id)aDelegate {
	return [[[RGraphicsDevice alloc] initWithContext:aContext deviceDesc:dd delegate:aDelegate] autorelease];
}

#undef deviceNumber
@synthesize scaleFactor,pointSize,textScale,deviceNumber;
#define Rf_deviceNumber deviceNumber

- (void)_updateDeviceParams {
	NewDevDesc *dev = (NewDevDesc*)devDesc;
	QuartzDesc *xd  = (QuartzDesc*)(dev->deviceSpecific);

	dev->cra[0] = 0.9*scaleFactor*pointSize;
	dev->cra[1] = 1.2*scaleFactor*pointSize;
	dev->ipr[0] = 1.0/(72.0*scaleFactor);
	dev->ipr[1] = 1.0/(72.0*scaleFactor);
	if(NULL != xd) xd->tscale = textScale;
}

- (void)setScaleFactor:(double)newScaleFactor {
	scaleFactor = newScaleFactor;
	[self _updateDeviceParams];
}

- (void)setPointSize:(double)newPointSize {
	if(newPointSize < 6.0) 	newPointSize = 6.0;
	pointSize = newPointSize;
	[self _updateDeviceParams];
}

- (void)setTextScale:(double)newTextScale {
	textScale = newTextScale;
	[self _updateDeviceParams];
}



- (NSSize)deviceSize {
	return [delegate graphicsDeviceSize:self];
}
//outerDevice... innerDevice... seems like a complex way to do
//something relatively simple.
- (void)createDevice:(NewDevDesc*)dev {
	dev->displayList = R_NilValue;
	
	dev->startfill = R_RGB(255,255,255);
	dev->startcol  = R_RGB(0,0,0);
	dev->startps   = pointSize;
	dev->startfont = 1;
	dev->startlty  = LTY_SOLID;
	dev->startgamma= 1;

	//Set up some happy pointers
	dev->newDevStruct = 1;
	dev->open         = RQuartz_Open;
	dev->close        = RQuartz_Close;
	dev->activate     = RQuartz_Activate;
	dev->deactivate   = RQuartz_Deactivate;
	dev->size         = RQuartz_Size;
	dev->newPage      = RQuartz_NewPage;
	dev->clip         = RQuartz_Clip;
	dev->strWidth     = RQuartz_StrWidth;
	dev->text         = RQuartz_Text;
	dev->rect         = RQuartz_Rect;
	dev->circle       = RQuartz_Circle;
	dev->line         = RQuartz_Line;
	dev->polyline     = RQuartz_Polyline;
	dev->polygon      = RQuartz_Polygon;
	dev->locator      = RQuartz_Locator;
	dev->mode         = RQuartz_Mode;
	dev->hold         = RQuartz_Hold;
	dev->metricInfo   = RQuartz_MetricInfo;
	
	//Set up the size
	NSSize frame = [self deviceSize];
	dev->left   = 0;
	dev->top    = 0;
	dev->right  = frame.width - 0.0001;
	dev->bottom = frame.height - 0.0001;
//	NSLog(NSStringFromSize(frame));
	
	//Magic numbers from on high.
	dev->xCharOffset = 0.4900;
	dev->yCharOffset = 0.3333;
	dev->yLineBias   = 0.10; //This is .2 for PS/PDF devices...
	
	[self _updateDeviceParams];
	
	dev->canResizePlot  = [delegate graphicsDeviceCanResize:self] == YES ? TRUE : FALSE;
	dev->canChangeFont  = TRUE;
	dev->canRotateText  = TRUE;
	dev->canClip        = TRUE;
	dev->canHAdj        = 2; //Arbirarily good horizontal adjustment
	dev->canChangeGamma = TRUE;
	dev->displayListOn  = TRUE;
	
	QuartzDesc *qd = calloc(1,sizeof(QuartzDesc));
	qd->device = self;
	qd->ctx    = context;
	qd->tscale = 1.0;
	qd->locator= NO;
	dev->deviceSpecific = qd;
	//Maybe? we can hope anyway...
	deviceNumber = devNumber((DevDesc*)dev)+1;
}

- (id)initWithContext:(CGContextRef)aContext deviceDesc:(void*)dd delegate:(id)aDelegate {
	if(nil == [super init]) return nil;
	delegate = [RGraphicsDeviceDelegateProxy proxyWithDelegate:aDelegate];
	context  = aContext;
	devDesc  = dd;
	scaleFactor = 1.0;
	pointSize   = 12.0;
	textScale   = 1.0;

	[self createDevice:(NewDevDesc*)dd];
	locateSet = NO;
	return self;
}

- (CGContextRef)context {
	return context;
}

- (void)setContext:(CGContextRef)aContext {
	NewDevDesc *dev = devDesc;
	if(NULL != dev) {
		QuartzDesc *qd = dev->deviceSpecific;
		if(NULL != qd) {
			qd->ctx = aContext;
			context = aContext;
		}
	}
}

- (void)setSize:(NSSize)newSize {
	NewDevDesc *dev = devDesc;
	NSLog(@"Changing device size: %@",NSStringFromSize(newSize));
	dev->bottom = newSize.height;
	dev->right  = newSize.width;
}


- (void)redraw {
	//Just in case something happens here.
	[self deviceSize];
	GEplayDisplayList((GEDevDesc*)GetDevice(devNumber((DevDesc*)devDesc))); 
}

- (CGRect)bounds {
	NSSize size = [self deviceSize];
	return (CGRect){0,0,size.width,size.height};
}

- (void)beginUpdate {
	[delegate graphicsDeviceWillDraw:self];
}

- (void)endUpdate {
	[delegate graphicsDeviceDidDraw:self];
}

- (void)startLocate {
	[delegate graphicsDeviceWillLocate:self];
}

- (void)endLocate {
	[delegate graphicsDeviceDidLocate:self];
}

- (BOOL)canLocate {
	return [delegate graphicsDeviceCanLocate:self];
}

- (BOOL)isLocating { return ((QuartzDesc*)((NewDevDesc*)devDesc)->deviceSpecific)->locator; }
- (void)setLocatePoint:(NSPoint)p {
	if(NO == [self isLocating]) return;
	locateSet   = YES;
	locatePoint = p;
	[NSApp postEvent:[NSEvent otherEventWithType:NSApplicationDefined 
									 location:NSMakePoint(0,0)
								modifierFlags:0 
									timestamp:0 
								 windowNumber:0 
									  context:nil 
									  subtype:1337
										data1:0 
										data2:0] atStart:NO];	
}
- (void)stopLocating {
	if(NO == [self isLocating]) return;
	((QuartzDesc*)((NewDevDesc*)devDesc)->deviceSpecific)->locator = NO;
	locateSet = NO;
	[NSApp postEvent:[NSEvent otherEventWithType:NSApplicationDefined 
									 location:NSMakePoint(0,0)
								modifierFlags:0 
									timestamp:0 
								 windowNumber:0 
									  context:nil 
									  subtype:1337
										data1:0 
										data2:0] atStart:NO];		
}

- (NSPoint)getPoint {
	locateSet = NO;
	[NSApp run];
	return locatePoint;
}

- (void)showCoord:(NSPoint)rPoint atPoint:(NSPoint)aPoint {
	[delegate graphicsDevice:self showCoords:rPoint atPoint:aPoint];
}

- (void)newPage {
	[delegate graphicsDeviceNewPage:self];
}

- (void)closeDevice {
	[delegate graphicsDeviceClose:self];
}

@end

#pragma mark Font Support
#pragma mark -
CGFontRef RQuartz_FontVariant(CGFontRef source,Boolean bold,Boolean italic) {
	if(false == bold && false == italic) return NULL;

	int i;
	CGFontRef variant      = NULL;
	CFArrayRef variants    = CGFontCopyVariationAxes(source);
	if(NULL == variants) return NULL;
	
	CFMutableDictionaryRef newDict = CFDictionaryCreateMutable(kCFAllocatorDefault,2,&kCFTypeDictionaryKeyCallBacks,&kCFTypeDictionaryValueCallBacks);
	Boolean    shouldCreate = false;
	//Find the bold and italic variants and update them if possible
	NSLog(@"Variation Axes: %d",CFArrayGetCount(variants));
	for(i=0;i<CFArrayGetCount(variants);i++) {
		CFDictionaryRef dict = (CFDictionaryRef)CFArrayGetValueAtIndex(variants,i);
		NSLog([(NSDictionary*)dict description]);
		CFRetain(dict);
		CFStringRef axisName = NULL;
		//If we are bold and have a "wght" variation, apply it.
		if(true == bold && 
		   CFDictionaryGetValueIfPresent(dict,kCGFontVariationAxisName,(void*)&axisName) && 
		   kCFCompareEqualTo == CFStringCompare(axisName,CFSTR("wght"),0)) {
			CFDictionaryAddValue(newDict,CFSTR("wght"),CFDictionaryGetValue(dict,kCGFontVariationAxisMaxValue));
			shouldCreate = true;
		}
		if(true == italic && 
		   CFDictionaryGetValueIfPresent(dict,kCGFontVariationAxisName,(void*)&axisName) && 
		   kCFCompareEqualTo == CFStringCompare(axisName,CFSTR("slnt"),0)) {
			CFDictionaryAddValue(newDict,CFSTR("slnt"),CFDictionaryGetValue(dict,kCGFontVariationAxisMaxValue));
			shouldCreate = true;			
		}
		
		
		CFRelease(dict);
	}
	if(true == shouldCreate) {
		variant = CGFontCreateCopyWithVariations(source,newDict);
	}
	CFRelease(newDict);
	CFRelease(variants);
	return variant;
}
CFStringRef RQuartz_FindFont(int fontface,char *fontfamily) {
	SEXP ns,env,db,names;
	PROTECT_INDEX index;
	CFStringRef fontName = CFSTR("");
	PROTECT(ns = R_FindNamespace(ScalarString(mkChar("grDevices"))));
	PROTECT_WITH_INDEX(env = findVar(install(".Quartzenv"),ns),&index);
	if(TYPEOF(env) == PROMSXP)
		REPROTECT(env = eval(env,ns),index);
	PROTECT(db    = findVar(install(".Quartz.Fonts"),env));
	PROTECT(names = getAttrib(db,R_NamesSymbol));
	if(strlen(fontfamily)>0) {
		int i;
		for(i=0;i<length(names);i++)
			if(0 == strcmp(fontfamily,CHAR(STRING_ELT(names,i)))) break;
		if(i<length(names))
			fontName = CFStringCreateWithCString(kCFAllocatorDefault,CHAR(STRING_ELT(VECTOR_ELT(db,i),fontface)), kCFStringEncodingUTF8);
	}
	UNPROTECT(4);
	return fontName;
}



#pragma mark Graphics Device Implementation
#pragma mark -

#define DEVDESC NewDevDesc *dd
#define CTXDESC R_GE_gcontext*gc,NewDevDesc*dd

#define DEVSPEC QuartzDesc *xd = (QuartzDesc*)dd->deviceSpecific;CGContextRef ctx = xd->ctx
#define DRAWSPEC QuartzDesc *xd = (QuartzDesc*)dd->deviceSpecific;CGContextRef ctx = xd->ctx;if(NULL==ctx) { printf("Invalid Context\n");return; }

CGFontRef RQuartz_Font(CTXDESC) {
	int fontface = gc->fontface;
	CFMutableStringRef fontName = CFStringCreateMutable(kCFAllocatorDefault,0);
	if((gc->fontface == 5) || (strcmp(gc->fontfamily,"symbol") == 0)) 
		CFStringAppend(fontName,CFSTR("Symbol"));
	else {
		CFStringRef font = RQuartz_FindFont(gc->fontface,gc->fontfamily);
		if(CFStringGetLength(font)>0) {
			fontface = 1; //This is handled by the lookup process		
			CFStringAppend(fontName,font);
		}
		CFRelease(font);
	}
	if(CFStringGetLength(fontName) == 0) {
		NSString *def = [[NSUserDefaults standardUserDefaults] stringForKey:@"DefaultFontName"];
		CFStringAppend(fontName,(nil != def) ? (CFStringRef)def : CFSTR("Arial"));
	}
	if(fontface==2 || fontface == 4) {
		CFStringAppend(fontName,CFSTR(" Bold"));
	}
	if(fontface==3) {
		CFStringAppend(fontName,CFSTR(" Italic"));
	}
	CGFontRef  font    = CGFontCreateWithFontName(fontName);
	CFRelease(fontName);
	return font;
}

#define RQUARTZ_FILL   (1)
#define RQUARTZ_STROKE (1<<1)
#define RQUARTZ_LINE   (1<<2)
#define RQUARTZ_FONT   (1<<3)

void RQuartz_Set(CGContextRef ctx,R_GE_gcontext*gc,NSUInteger flags) { 
	if(flags & RQUARTZ_FILL) {
		int fill = gc->fill;
		CGContextSetRGBFillColor(ctx,R_RED(fill)/256.0,R_GREEN(fill)/256.0,R_BLUE(fill)/256.0,R_ALPHA(fill)/256.0);
	}
	if(flags & RQUARTZ_STROKE) {
		int stroke = gc->col;
		CGContextSetRGBStrokeColor(ctx,R_RED(stroke)/256.0,R_GREEN(stroke)/256.0,R_BLUE(stroke)/256.0,R_ALPHA(stroke)/256.0);
	}
	if(flags & RQUARTZ_LINE) {
		CGFloat dashlist[8];
		int   i,ndash = 0;
		int   lty = gc->lty;
		CGContextSetLineWidth(ctx,gc->lwd);
		
		CGFloat lwd = gc->lwd*0.75;
		for(i=0;i<8 && lty;i++) {
			dashlist[ndash++] = (lwd >= 1 ? lwd : 1)*(lty&15);
			lty >>= 4;
		}
		CGContextSetLineDash(ctx,0,dashlist,ndash);
		CGLineCap cap;
		switch(gc->lend) {
		case GE_ROUND_CAP:cap = kCGLineCapRound;break;
		case GE_BUTT_CAP:cap = kCGLineCapButt;break;
		case GE_SQUARE_CAP:cap = kCGLineCapSquare;break;
		}
		CGContextSetLineCap(ctx,cap);
		CGLineJoin join;
		switch(gc->ljoin) {
		case GE_ROUND_JOIN:join = kCGLineJoinRound;break;
		case GE_MITRE_JOIN:join = kCGLineJoinMiter;break;
		case GE_BEVEL_JOIN:join = kCGLineJoinBevel;break;
		}
		CGContextSetLineJoin(ctx,join);
		CGContextSetMiterLimit(ctx,gc->lmitre);
	}
	if(flags & RQUARTZ_FONT) {
		CGFontRef font = RQuartz_Font(gc,NULL);
		CGContextSetFont(ctx,font);
		CGContextSetFontSize(ctx,gc->cex*gc->ps);
	}
}

#define SET(X) RQuartz_Set(ctx,gc,(X))

static Rboolean RQuartz_Open(DEVDESC,QuartzDesc *xd,char *display,double width,double height,int bg) {
	//We don't do anything here.
	return TRUE;
}

static void RQuartz_Close(DEVDESC) {
	DRAWSPEC;
	[xd->device closeDevice];	
}

static void RQuartz_Activate(DEVDESC) {
	//And here
}

static void RQuartz_Deactivate(DEVDESC) {
	//And here.
}

static void RQuartz_Size(double*left,double*right,double*bottom,double*top,DEVDESC) {
	DRAWSPEC;
	NSSize s = [xd->device deviceSize];
	*left = *top = 0;
	*right = s.width;
	*bottom= s.height;
}

static void RQuartz_NewPage(CTXDESC) {
	DRAWSPEC;
	[xd->device newPage];
	RQuartz_Mode(1,dd);
	SET(RQUARTZ_FILL);
	CGRect bounds = [xd->device bounds];
	CGContextFillRect(ctx,bounds);
	RQuartz_Mode(0,dd);
}

static void RQuartz_Clip(double x0,double x1,double y0,double y1,DEVDESC) {
}

extern CGFontRef CGContextGetFont(CGContextRef);

CFStringRef prepareText(CTXDESC,char *text,UniChar **buffer,int *free) {
	CFStringRef str;
	if(gc->fontface == 5 || strcmp(gc->fontfamily,"symbol") == 0)
		str = CFStringCreateWithCString(NULL,text,kCFStringEncodingMacSymbol);
	else {
		str = CFStringCreateWithCString(NULL,text,kCFStringEncodingUTF8);
		//Try fallback string encodings if UTF8 doesn't work.
		if(NULL == str)
			CFStringCreateWithCString(NULL,text,kCFStringEncodingISOLatin1);
	}
//	NSLog(@"prepareText: %@",str);
	*buffer = (UniChar*)CFStringGetCharactersPtr(str);
	if (*buffer == NULL) {
		CFIndex length = CFStringGetLength(str);
		*buffer = malloc(length * sizeof(UniChar));
		CFStringGetCharacters(str, CFRangeMake(0, length), *buffer);
		*free = 1;
	}
	return str;
}

static double RQuartz_StrWidth(char *text,CTXDESC) {
	DEVSPEC;
	if(ctx==NULL) return 0.0;
	SET(RQUARTZ_FONT);
	CGFloat aScale   = (gc->cex*gc->ps)/(CGFontGetUnitsPerEm(CGContextGetFont(ctx)));
	UniChar *buffer;
	CGGlyph   *glyphs;
	int      *advances;
	int Free = 0,len,i;
	CFStringRef str = prepareText(gc,dd,text,&buffer,&Free);
	len = CFStringGetLength(str);
	glyphs = malloc(sizeof(CGGlyph)*len);
	advances = malloc(sizeof(int)*len);
	CGFontGetGlyphsForUnichars(CGContextGetFont(ctx),buffer,glyphs,len);
	CGFontGetGlyphAdvances(CGContextGetFont(ctx),glyphs,len,advances);
	CGFloat width = 0.0;// aScale*CGFontGetLeading(CGContextGetFont(ctx));
	for(i=0;i<len;i++) width += aScale*advances[i];
	free(advances);
	free(glyphs);
	if(Free) free(buffer);
	CFRelease(str);
	return width;
}

static void RQuartz_Text(double x,double y,char *text,double rot,double hadj,CTXDESC) {
	DRAWSPEC;
	//A stupid hack because R isn't consistent.
	int fill = gc->fill;
	gc->fill = gc->col;
	SET(RQUARTZ_FILL|RQUARTZ_STROKE|RQUARTZ_FONT);
	gc->fill = fill;
	CGFontRef font = CGContextGetFont(ctx);
	CGFloat aScale   = (gc->cex*gc->ps)/(CGFontGetUnitsPerEm(font));
	UniChar *buffer;
	CGGlyph   *glyphs;

	int Free = 0,len,i;
	CGFloat width = 0.0;
	CFStringRef str = prepareText(gc,dd,text,&buffer,&Free);
	len = CFStringGetLength(str);
	glyphs = malloc(sizeof(CGGlyph)*len);
	CGFontGetGlyphsForUnichars(font,buffer,glyphs,len);
	int      *advances = malloc(sizeof(int)*len);
	CGSize   *g_adv    = malloc(sizeof(CGSize)*len);
	
	CGFontGetGlyphAdvances(font,glyphs,len,advances);
	for(i=0;i<len;i++) { width += advances[i]*aScale;g_adv[i] = CGSizeMake(aScale*advances[i]*cos(-0.0174532925*rot),aScale*advances[i]*sin(-0.0174532925*rot)); }
	free(advances);
	CGContextSetTextMatrix(ctx,CGAffineTransformConcat(CGAffineTransformMakeScale(xd->tscale,-xd->tscale),
													   CGAffineTransformMakeRotation(-0.0174532925*rot)));
	double ax = (width*hadj)*cos(-0.0174532925*rot);
	double ay = (width*hadj)*sin(-0.0174532925*rot);
//	double h  = CGFontGetXHeight(CGContextGetFont(ctx))*aScale;
	CGContextSetTextPosition(ctx,x-ax,y-ay);
//	Rprintf("%s,%.2f %.2f (%.2f,%.2f) (%d,%f)\n",text,hadj,width,ax,ay,CGFontGetUnitsPerEm(CGContextGetFont(ctx)),CGContextGetFontSize(ctx));	
	CGContextShowGlyphsWithAdvances(ctx,glyphs,g_adv,len);
	free(glyphs);
	free(g_adv);
	if(Free) free(buffer);
	CFRelease(str);
}

static void RQuartz_Rect(double x0,double y0,double x1,double y1,CTXDESC) {
	DRAWSPEC;
	SET(RQUARTZ_FILL|RQUARTZ_STROKE|RQUARTZ_LINE);
	CGContextBeginPath(ctx);
	CGContextAddRect(ctx,CGRectMake(x0,y0,x1-x0,y1-y0));
	CGContextDrawPath(ctx,kCGPathFillStroke);
}

static void RQuartz_Circle(double x,double y,double r,CTXDESC) {
	DRAWSPEC;
	SET(RQUARTZ_FILL|RQUARTZ_STROKE|RQUARTZ_LINE);
	double r2 = 2.0*r;
	CGContextBeginPath(ctx);
	CGContextAddEllipseInRect(ctx,CGRectMake(x-r,y-r,r2,r2));
	CGContextDrawPath(ctx,kCGPathFillStroke);
}

static void RQuartz_Line(double x1,double y1,double x2,double y2,CTXDESC) {
	DRAWSPEC;
	SET(RQUARTZ_STROKE|RQUARTZ_LINE);
	CGContextBeginPath(ctx);
	CGContextMoveToPoint(ctx,x1,y1);
	CGContextAddLineToPoint(ctx,x2,y2);
	CGContextStrokePath(ctx);
}

static void RQuartz_Polyline(int n,double *x,double *y,CTXDESC) {
	if(n<2) return;	
	int i;
	DRAWSPEC;
	SET(RQUARTZ_STROKE|RQUARTZ_LINE);	
	CGContextBeginPath(ctx);
	CGContextMoveToPoint(ctx,x[0],y[0]);
	for(i=1;i<n;i++) CGContextAddLineToPoint(ctx,x[i],y[i]);
	CGContextStrokePath(ctx);
}

static void RQuartz_Polygon(int n,double *x,double *y,CTXDESC) {
	if(n<2) return;	
	int i;
	DRAWSPEC;
	SET(RQUARTZ_FILL|RQUARTZ_STROKE|RQUARTZ_LINE);	
	CGContextBeginPath(ctx);
	CGContextMoveToPoint(ctx,x[0],y[0]);
	for(i=1;i<n;i++) CGContextAddLineToPoint(ctx,x[i],y[i]);
	CGContextClosePath(ctx);
	CGContextDrawPath(ctx,kCGPathFillStroke);	
}

static void RQuartz_Mode(int mode,DEVDESC) {
	DEVSPEC;
	if(0 == mode) { CGContextSynchronize(ctx);[xd->device beginUpdate]; } else { [xd->device endUpdate]; }
}

static void RQuartz_Hold(DEVDESC) {
}

static void RQuartz_MetricInfo(int c,R_GE_gcontext *gc,double *ascent,double *descent,double *width,NewDevDesc *dd) {
	DRAWSPEC;
	SET(RQUARTZ_FONT);
	char    text[2];
	text[0] = c;
	text[1] = '\0';
	CGFontRef font = CGContextGetFont(ctx);
	CGFloat aScale   = (gc->cex*gc->ps)/CGFontGetUnitsPerEm(font);
	UniChar *buffer;
	CGGlyph   *glyphs;
	int Free = 0,len,i;
	*width = 0.0;
	CFStringRef str = prepareText(gc,dd,text,&buffer,&Free);
	len = CFStringGetLength(str);
	glyphs = malloc(sizeof(CGGlyph)*len);
	CGFontGetGlyphsForUnichars(font,buffer,glyphs,len);
	int      *advances = malloc(sizeof(int)*len);
	CGRect   *bboxes   = malloc(sizeof(CGRect)*len);
	CGFontGetGlyphAdvances(font,glyphs,len,advances);
	CGFontGetGlyphBBoxes(font,glyphs,len,bboxes);
	for(i=0;i<len;i++) *width += advances[i]*aScale;
	*ascent  = aScale*(bboxes[0].size.height + bboxes[0].origin.y);
	*descent = -aScale*bboxes[0].origin.y;
	free(bboxes);
	free(advances);
	free(glyphs);
	if(Free) free(buffer);
	CFRelease(str);
}

static Rboolean RQuartz_Locator(double *x,double *y,DEVDESC) {
	DEVSPEC;
	ctx = NULL;
	if(NO == [xd->device canLocate])
		return FALSE;
	if(NO == xd->locator) {
		xd->locator = YES;		
		[xd->device startLocate];
	}
	NSPoint p = [xd->device getPoint];	
	if(YES == xd->locator) {
		*x = p.x;
		*y = p.y;
		//NSMakePoint(GConvertX(p.x,0,12,(DevDesc*)dd),GConvertY(p.y,0,12,(DevDesc*)dd))
		[xd->device showCoord:p atPoint:p];
		return TRUE;
	} else {
		[xd->device endLocate];
		return FALSE;
	}
}


